package meerkat.voting.output;

import com.google.common.util.concurrent.FutureCallback;
import com.google.protobuf.ByteString;
import meerkat.protobuf.Crypto.*;
import meerkat.protobuf.Voting.*;
import meerkat.voting.controller.callbacks.ControllerCallback;
import meerkat.voting.output.outputcommands.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.LinkedBlockingQueue;

/**
 * A toy OutputDevice class
 * outputs everything simply to the System console
 */
public class SystemConsoleOutputDevice implements BallotOutputDevice, Runnable {

    private Logger logger;
    private LinkedBlockingQueue<OutputCommand> queue;
    private volatile boolean shutDownHasBeenCalled;

    public SystemConsoleOutputDevice () {
        logger = LoggerFactory.getLogger(SystemConsoleOutputDevice.class);
        logger.info("A SystemConsoleOutputDevice is constructed");
        queue = new LinkedBlockingQueue<>();
        shutDownHasBeenCalled = false;
    }


    /*
     * Returns the UTF8 decoding of byte-string data
     */
    private static String bytesToString(ByteString data) {
        return data.toStringUtf8();
    }

    @Override
    public void run () {
        logger.info("UI starts running");
        while (! wasShutDownCalled()) {
            try {
                OutputCommand command = queue.take();
                handleSingleCommand(command);
            }
            catch (InterruptedException e) {
                logger.warn ("Interrupted while reading from command queue " + e);
            }
        }
    }

    private boolean wasShutDownCalled () {
        return shutDownHasBeenCalled;
    }

    @Override
    public void callShutDown() {
        logger.info("callShutDown command has been called");
        shutDownHasBeenCalled = true;
        queue.clear();
    }


    private void handleSingleCommand(OutputCommand command) {
        if (command instanceof CommitOutputCommand) {
            doCommitToBallot((CommitOutputCommand)command);
        }
        else if (command instanceof AuditOutputCommand) {
            doAudit((AuditOutputCommand)command);
        }
        else if (command instanceof CastOutputCommand) {
            doCastBallot((CastOutputCommand)command);
        }
        else if (command instanceof CancelOutputCommand) {
            doCancel((CancelOutputCommand)command);
        }
        else {
            String errorMessage = "handleSingleCommand: unknown type of OutputCommand received: " +
                    command.getClass().getName();
            logger.error(errorMessage);
            throw new RuntimeException(errorMessage);
        }
    }


    @Override
    public void commitToBallot(PlaintextBallot plaintextBallot,
                               SignedEncryptedBallot signedEncryptedBallot,
                               FutureCallback<Void> callback) {
        logger.debug("Output interface call to commit to ballot");
        queue.add(new CommitOutputCommand(plaintextBallot, signedEncryptedBallot, (ControllerCallback)callback));
    }

    public void doCommitToBallot(CommitOutputCommand command) {
        logger.debug("entered method doCommitToBallot");
        PlaintextBallot plaintextBallot = command.getPlaintext();
        long plaintextSerialNumber = plaintextBallot.getSerialNumber();
        System.out.println("Commitment of Ballot #" + plaintextSerialNumber + " (plaintext):");
        System.out.println(plaintextBallot);
        SignedEncryptedBallot signedEncryptedBallot = command.getSignedEncryptedBallot();
        long encryptedSerialNumber = signedEncryptedBallot.getEncryptedBallot().getSerialNumber();
        System.out.println("Commitment of Ballot #" + encryptedSerialNumber + " (ciphertext):");
        if (plaintextSerialNumber != encryptedSerialNumber) {
            logger.error("plaintext and encryption serial numbers do not match!! plaintext# = " +
                    plaintextSerialNumber + ", ciphertext# = " + encryptedSerialNumber);
        }
        ByteString encryptedData = signedEncryptedBallot.getEncryptedBallot().getData().getData();
        System.out.println(bytesToString(encryptedData));
        command.getCallback().onSuccess(null);
    }


    @Override
    public void audit(BallotSecrets ballotSecrets, FutureCallback<Void> callback) {
        logger.debug("an interface call to audit");
        queue.add(new AuditOutputCommand(ballotSecrets, (ControllerCallback)callback));
    }

    public void doAudit(AuditOutputCommand command) {
        logger.debug("entered method doAudit");
        System.out.println("Auditing");
        BallotSecrets ballotSecrets = command.getBallotSecrets();
        printEncryptionRandomness (ballotSecrets.getEncryptionRandomness());
        printRandomnessGenerationProof (ballotSecrets.getProof());
        command.getCallback().onSuccess(null);
    }


    @Override
    public void castBallot(FutureCallback<Void> callback) {
        logger.debug("an interface call to cast ballot");
        queue.add(new CastOutputCommand((ControllerCallback)callback));
    }

    public void doCastBallot(CastOutputCommand command) {
        logger.debug("entered method doCastBallot");
        System.out.println("Ballot finalized for casting!");
        command.getCallback().onSuccess(null);
    }


    @Override
    public void cancelBallot(FutureCallback<Void> callback) {
        logger.debug("an interface call to cancel the output");
        queue.add(new CancelOutputCommand((ControllerCallback)callback));
    }

    public void doCancel(CancelOutputCommand command) {
        logger.debug("entered method doCancel");
        System.out.println("Ballot cancelled!");
        command.getCallback().onSuccess(null);
    }


    private void printEncryptionRandomness (EncryptionRandomness encryptionRandomness) {
        System.out.println("Encryption Randomness = ");
        ByteString data = encryptionRandomness.getData();
        System.out.println(bytesToString(data));
    }

    private void printRandomnessGenerationProof (RandomnessGenerationProof proof) {
        System.out.println("Proof of randomness generation:");
        ByteString data = proof.getData();
        System.out.println(bytesToString(data));
    }

}
